///////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2010 Sebastian Pancratz                                     //
//                                                                           //
// Distributed under the terms of the GNU General Public License as          //
// published by the Free Software Foundation; either version 2 of the        //
// License, or (at your option) any later version.                           //
//                                                                           //
// http://www.gnu.org/licenses/                                              //
///////////////////////////////////////////////////////////////////////////////

#ifndef _FMPQ_POLY_H_
#define _FMPQ_POLY_H_

#include <stdio.h>
#include <gmp.h>

#include "fmpz.h"
#include "fmpz_poly.h"

/**
 * \file     fmpq_poly.h
 * \brief    Fast implementation of the rational polynomial ring
 * \author   Sebastian Pancratz
 * \date     Mar 2010 -- July 2010
 * \version  0.1.2
 */

/**
 * \ingroup  Definitions
 * \brief    Data type for a rational polynomial.
 * \details  We represent a rational polynomial as the quotient of an integer 
 *           polynomial of type <tt>fmpz_poly_t</tt> and a denominator of type 
 *           <tt>fmpz_t</tt>, enforcing coprimality and that the denominator 
 *           is positive.  Moreover, the denominator may be <tt>NULL</tt>, 
 *           which is interpreted as \f$1\f$.  The zero polynomial is 
 *           represented as \f$0/1\f$.
 */
typedef struct
{
    fmpz_poly_p num; //!< Numerator
    fmpz_t den;      //!< Denominator
} fmpq_poly_struct;

/**
 * \ingroup  Definitions
 * \brief    Array of #fmpq_poly_struct of length one.
 * \details  This allows passing <em>by reference</em> without having to 
 *           explicitly allocate memory for the structure, as one would have 
 *           to when using pointers.
 */
typedef fmpq_poly_struct fmpq_poly_t[1];

/**
 * \ingroup  Definitions
 * \brief    Pointer to #fmpq_poly_struct.
 */
typedef fmpq_poly_struct * fmpq_poly_ptr;

void fmpq_poly_canonicalize(fmpq_poly_ptr f, fmpz_t temp);

///////////////////////////////////////////////////////////////////////////////
// _fmpq_poly_*                                                              //
///////////////////////////////////////////////////////////////////////////////

/**
 * \ingroup  Helper
 * 
 * Ensures that <tt>op-\>den</tt> can hold <tt>limbs</tt> limbs (plus a 
 * sign/ size limb).
 * 
 * This method never shrinks <tt>op-\>den</tt>.  Moreover, after a call to 
 * this method, <tt>op-\>den</tt> is guaranteed to be initialised.
 */
static inline
void _fmpq_poly_den_fit_limbs(fmpq_poly_ptr op, unsigned long limbs)
{
    if (op->den == NULL)
    {
        op->den = fmpz_init(limbs > 1 ? limbs : 1);
        fmpz_set_si(op->den, 1);  /* XXX: FLINT, Valgrind problem */
    }
    else if (fmpz_size(op->den) < limbs)
        op->den = fmpz_realloc(op->den, limbs);
}

/**
 * \ingroup  Helper
 * 
 * Compares the denominators of <tt>op1</tt> and <tt>op2</tt>.
 * 
 * This method handles <tt>NULL</tt> values correctly.  This method only deals 
 * with the <tt>den</tt> attributes of the arguments.  In particular, it does 
 * not ensure that the contents of the numerators are coprime to the 
 * denominators.
 * 
 * All arguments may be aliased.
 */
static inline
int _fmpq_poly_den_equal(fmpq_poly_ptr op1, fmpq_poly_ptr op2)
{
    if (op1->den == NULL)
        return op2->den == NULL || fmpz_is_one(op2->den);
    else
        return op2->den == NULL ? fmpz_is_one(op1->den) 
                                : fmpz_equal(op1->den, op2->den);
}

/**
 * \ingroup  Helper
 * 
 * Returns whether the denominator of <tt>op</tt> is one.
 * 
 * This method does not check that the denominator is coprime to the 
 * numerator.
 */
static inline
int _fmpq_poly_den_is_one(fmpq_poly_ptr op)
{
    return op->den == NULL || fmpz_is_one(op->den);
}

///////////////////////////////////////////////////////////////////////////////
// fmpq_poly_*                                                               //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Accessing numerator and denominator

/**
 * \def      fmpq_poly_numref(op)
 * \brief    Returns a reference to the numerator of \c op.
 * \ingroup  NumDen
 * \details  The <tt>fmpz_poly</tt> functions can be used on the result of 
 *           this macro.
 */
#define fmpq_poly_numref(op)  ((op)->num)

/**
 * \def      fmpq_poly_denref(op)
 * \brief    Returns a reference to the denominator of \c op.
 * \ingroup  NumDen
 * \details  The <tt>fmpz</tt> functions can be used on the result of 
 *           this macro \e if the result is not <tt>NULL</tt>.
 */
#define fmpq_poly_denref(op)  ((op)->den)

/**
 * \brief    Sets \c rop to the numerator of \c op.
 * \ingroup  NumDen
 * 
 * \note  The arguments <tt>rop</tt> and <tt>op-\>num</tt> may be aliased.
 * 
 * \param[out] rop  Integer polynomial
 * \param[in] op    Rational polynomial
 */
static inline 
void fmpq_poly_get_num(fmpz_poly_p rop, const fmpq_poly_ptr op)
{
    fmpz_poly_set(rop, op->num);
}

/**
 * \brief    Sets \c rop to the denominator of \c op.
 * \ingroup  NumDen
 * 
 * This method handles <tt>NULL</tt> values correctly.
 * 
 * \note  The arguments <tt>rop</tt> and <tt>op-\>den</tt> may be aliased.
 * 
 * \param[out] rop  Integer
 * \param[in] op    Rational polynomial
 */
static inline 
void fmpq_poly_get_den(fmpz_t rop, const fmpq_poly_ptr op)
{
    if (rop == op->den)  /* Aliasing */
        return;
    
    if (op->den == NULL)
        fmpz_set_si(rop, 1);
    else
    {
        fmpz_set_si(rop, 1);
        if (fmpz_size(rop) < fmpz_size(op->den))
            rop = fmpz_realloc(rop, fmpz_size(op->den));
        fmpz_set(rop, op->den);
    }
}

/**
 * \brief    Sets the numerator of \c rop to the polynomial \c op. 
 * \ingroup  NumDen
 * 
 * \note  The arguments <tt>rop-\>num</tt> and <tt>op</tt> may be aliased.
 * 
 * \param[out] rop  Rational polynomial
 * \param[in] op    Integer polynomial
 */
static inline 
void fmpq_poly_set_num(fmpq_poly_ptr rop, const fmpz_poly_p op)
{
    fmpz_poly_set(rop->num, op);
}

/**
 * \brief    Sets the denominator of \c rop to the integer \c op. 
 * \ingroup  NumDen
 * 
 * This method handles <tt>NULL</tt> values correctly.  It does \e not ensure 
 * that \c rop ends up in canonical form.
 * 
 * The arguments <tt>rop-\>den</tt> and <tt>op</tt> may be aliased.
 * 
 * \param[out] rop  Rational polynomial
 * \param[in] op    Integer
 */
static inline 
void fmpq_poly_set_den(fmpq_poly_ptr rop, const fmpz_t op)
{
    if (op == NULL)
    {
        if (rop->den != NULL)
            fmpz_set_si(rop->den, 1);
    }
    else
    {
        if (rop->den == NULL)
        {
            if (!fmpz_is_one(op))
            {
                rop->den = fmpz_init(fmpz_size(op));
                fmpz_set(rop->den, op);
            }
        }
        else
        {
            if (fmpz_size(rop->den) < fmpz_size(op))
                rop->den = fmpz_realloc(rop->den, fmpz_size(op));
            fmpz_set(rop->den, op);
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
// Memory management

/**
 * \ingroup  MemoryManagement
 * 
 * Initializes the element \c rop to zero.
 * 
 * This function should be called once for the #fmpq_poly_ptr \c rop, before 
 * using it with other <tt>fmpq_poly</tt> functions, or following a 
 * preceeding call to #fmpq_poly_clear().
 */
static inline 
void fmpq_poly_init(fmpq_poly_ptr rop)
{
    rop->num = (fmpz_poly_p) malloc(sizeof(fmpz_poly_struct));
    fmpz_poly_init(rop->num);
    rop->den = NULL;
}

/**
 * \ingroup  MemoryManagement
 *
 * Clears the element \c rop.
 * 
 * This function should only be called on an element which has been 
 * initialised.
 */
static inline 
void fmpq_poly_clear(fmpq_poly_ptr rop)
{
    if (rop->num != NULL)
    {
        fmpz_poly_clear(rop->num);
        free(rop->num);
        rop->num = NULL;
    }
    if (rop->den != NULL);
    {
        fmpz_clear(rop->den);
        rop->den = NULL;
    }
}

///////////////////////////////////////////////////////////////////////////////
// Assignment and basic manipulation

/**
 * \ingroup  Assignment
 * 
 * Sets the element \c rop to the same value as the element \c op.
 */
static inline 
void fmpq_poly_set(fmpq_poly_ptr rop, const fmpq_poly_ptr op)
{
    if (rop != op)
    {
        fmpz_poly_set(rop->num, op->num);
        if (op->den == NULL)
        {
            if (rop->den != NULL)
            {
                fmpz_clear(rop->den);
                rop->den = NULL;
            }
        }
        else
        {
            _fmpq_poly_den_fit_limbs(rop, fmpz_size(op->den));
            fmpz_set(rop->den, op->den);
        }
    }
}

/**
 * \ingroup  Assignment
 * 
 * Sets the element \c rop to the value given by the \c int \c op.
 */
static inline  
void fmpq_poly_set_si(fmpq_poly_ptr rop, const long op)
{
    fmpz_poly_zero(rop->num);
    fmpz_poly_set_coeff_si(rop->num, 0, op);
    if (rop->den != NULL)
        fmpz_set_si(rop->den, 1);
}

/**
 * \ingroup  Assignment
 * 
 * Sets the element \c rop to the integer \c x.
 */
static inline
void fmpq_poly_set_fmpz(fmpq_poly_ptr rop, const fmpz_t x)
{
    fmpz_poly_zero(rop->num);
    fmpz_poly_set_coeff_fmpz(rop->num, 0, x);
    if (rop->den != NULL)
        fmpz_set_si(rop->den, 1);
}

/**
 * \ingroup  Assignment
 * 
 * Sets the element \c rop to the integer \c x.
 */
static inline
void fmpq_poly_set_mpz(fmpq_poly_ptr rop, const mpz_t x)
{
    fmpz_poly_zero(rop->num);
    fmpz_poly_set_coeff_mpz(rop->num, 0, x);
    if (rop->den != NULL)
        fmpz_set_si(rop->den, 1);
}

/**
 * \ingroup  Assignment
 * 
 * Sets the element \c rop to the rational \c x.
 */
static inline
void fmpq_poly_set_mpq(fmpq_poly_ptr rop, const mpq_t x)
{
    fmpz_poly_zero(rop->num);
    fmpz_poly_set_coeff_mpz(rop->num, 0, mpq_numref(x));
    if (mpz_cmp_si(mpq_denref(x), 1))  /* x->den > 1 */
    {
        _fmpq_poly_den_fit_limbs(rop, mpz_size(mpq_denref(x)));
        mpz_to_fmpz(rop->den, mpq_denref(x));
    }
    else                               /* x.den == 1 */
        if (rop->den != NULL)
            fmpz_set_si(rop->den, 1);
}

/**
 * \ingroup  Assignment
 * 
 * Swaps the elements \c op1 and \c op2.
 * 
 * This is done efficiently by swapping pointers.
 */
static inline 
void fmpq_poly_swap(fmpq_poly_ptr op1, fmpq_poly_ptr op2)
{
    fmpz_t t;
    if (op1 != op2)
    {
        fmpz_poly_swap(op1->num, op2->num);
        t        = op1->den;
        op1->den = op2->den;
        op2->den = t;
    }
}

/**
 * \ingroup  Assignment
 * 
 * Sets the element \c rop to zero.
 */
static inline
void fmpq_poly_zero(fmpq_poly_ptr rop)
{
    fmpz_poly_zero(rop->num);
    if (rop->den != NULL)
        fmpz_set_si(rop->den, 1);
}

/**
 * \ingroup  Assignment
 * 
 * Sets the element \c rop to one.
 */
static inline
void fmpq_poly_one(fmpq_poly_ptr rop)
{
    fmpz_poly_zero(rop->num);
    fmpz_poly_set_coeff_si(rop->num, 0, 1);
    if (rop->den != NULL)
        fmpz_set_si(rop->den, 1);
}

void fmpq_poly_neg(fmpq_poly_ptr rop, const fmpq_poly_ptr op);
void fmpq_poly_inv(fmpq_poly_ptr rop, const fmpq_poly_ptr op);

///////////////////////////////////////////////////////////////////////////////
// Setting/ retrieving individual coefficients

void fmpq_poly_get_coeff_mpq(mpq_t rop, const fmpq_poly_ptr op, unsigned long i);

void fmpq_poly_set_coeff_fmpz(fmpq_poly_ptr rop, const unsigned long i, const fmpz_t x);
void fmpq_poly_set_coeff_mpz(fmpq_poly_ptr rop, const unsigned long i, const mpz_t x);
void fmpq_poly_set_coeff_mpq(fmpq_poly_ptr rop, const unsigned long i, const mpq_t x);
void fmpq_poly_set_coeff_si(fmpq_poly_ptr rop, const unsigned long i, long x);

///////////////////////////////////////////////////////////////////////////////
// Comparison

/**
 * \brief    Returns whether \c op is zero.
 * \ingroup  Comparison
 * 
 * Returns whether the element \c op is zero.
 */
static inline
int fmpq_poly_is_zero(const fmpq_poly_ptr op)
{
    return fmpz_poly_degree(op->num) < 0;
}

/**
 * \brief    Returns whether \c op is equal to \f$1\f$.
 * \ingroup  Comparison
 * 
 * Returns whether the element \c op is equal to the constant polynomial 
 * \f$1\f$.
 */
static inline
int fmpq_poly_is_one(const fmpq_poly_ptr op)
{
    return (op->num->length == 1) && fmpz_is_one(op->num->coeffs) 
        && _fmpq_poly_den_is_one(op);
}

int fmpq_poly_equal(const fmpq_poly_ptr op1, const fmpq_poly_ptr op2);
int fmpq_poly_cmp(const fmpq_poly_ptr left, const fmpq_poly_ptr right);

///////////////////////////////////////////////////////////////////////////////
// Polynomial parameters

/**
 * \brief    Returns the length of \c op.
 * \ingroup  PolyParameters
 * \details  Returns the length of the polynomial \c op, which is one greater 
 *           than its degree.
 */
static inline 
unsigned long fmpq_poly_length(const fmpq_poly_ptr op)
{
    return fmpz_poly_length(op->num);
}

/**
 * \brief    Returns the degree of \c op.
 * \ingroup  Parameters
 * \details  Returns the degree of the polynomial \c op.
 */
static inline
long fmpq_poly_degree(const fmpq_poly_ptr op)
{
    return fmpz_poly_degree(op->num);
}

///////////////////////////////////////////////////////////////////////////////
// Addition/ subtraction

void fmpq_poly_add(fmpq_poly_ptr rop, const fmpq_poly_ptr op1, const fmpq_poly_ptr op2);
void fmpq_poly_sub(fmpq_poly_ptr rop, const fmpq_poly_ptr op1, const fmpq_poly_ptr op2);

void fmpq_poly_addmul(fmpq_poly_ptr rop, const fmpq_poly_ptr op1, const fmpq_poly_ptr op2);
void fmpq_poly_submul(fmpq_poly_ptr rop, const fmpq_poly_ptr op1, const fmpq_poly_ptr op2);

///////////////////////////////////////////////////////////////////////////////
// Scalar multiplication and division

void fmpq_poly_scalar_mul_si(fmpq_poly_ptr rop, const fmpq_poly_ptr op, long x);
void fmpq_poly_scalar_mul_mpz(fmpq_poly_ptr rop, const fmpq_poly_ptr op, const mpz_t x);
void fmpq_poly_scalar_mul_mpq(fmpq_poly_ptr rop, const fmpq_poly_ptr op, const mpq_t x);

void fmpq_poly_scalar_div_si(fmpq_poly_ptr rop, const fmpq_poly_ptr op, long x);
void fmpq_poly_scalar_div_mpz(fmpq_poly_ptr rop, const fmpq_poly_ptr op, const mpz_t x);
void fmpq_poly_scalar_div_mpq(fmpq_poly_ptr rop, const fmpq_poly_ptr op, const mpq_t x);

///////////////////////////////////////////////////////////////////////////////
// Multiplication

void fmpq_poly_mul(fmpq_poly_ptr rop, const fmpq_poly_ptr op1, const fmpq_poly_ptr op2);

///////////////////////////////////////////////////////////////////////////////
// Division

void fmpq_poly_floordiv(fmpq_poly_ptr q, const fmpq_poly_ptr a, const fmpq_poly_ptr b);
void fmpq_poly_mod(fmpq_poly_ptr r, const fmpq_poly_ptr a, const fmpq_poly_ptr b);
void fmpq_poly_divrem(fmpq_poly_ptr q, fmpq_poly_ptr r, const fmpq_poly_ptr a, const fmpq_poly_ptr b);

///////////////////////////////////////////////////////////////////////////////
// Powering

void fmpq_poly_power(fmpq_poly_ptr rop, const fmpq_poly_ptr op, unsigned long exp);

///////////////////////////////////////////////////////////////////////////////
// Greatest common divisor

void fmpq_poly_gcd(fmpq_poly_ptr rop, const fmpq_poly_ptr a, const fmpq_poly_ptr b);
void fmpq_poly_xgcd(fmpq_poly_ptr rop, fmpq_poly_ptr s, fmpq_poly_ptr t, const fmpq_poly_ptr a, const fmpq_poly_ptr b);
void fmpq_poly_lcm(fmpq_poly_ptr rop, const fmpq_poly_ptr a, const fmpq_poly_ptr b);

///////////////////////////////////////////////////////////////////////////////
// Derivative

void fmpq_poly_derivative(fmpq_poly_ptr rop, fmpq_poly_ptr op);

///////////////////////////////////////////////////////////////////////////////
// Evaluation

void fmpq_poly_evaluate_mpz(mpq_t rop, fmpq_poly_ptr f, const mpz_t a);
void fmpq_poly_evaluate_mpq(mpq_t rop, fmpq_poly_ptr f, const mpq_t a);

///////////////////////////////////////////////////////////////////////////////
// Gaussian content

void fmpq_poly_content(mpq_t rop, const fmpq_poly_ptr op);
void fmpq_poly_primitive_part(fmpq_poly_ptr rop, const fmpq_poly_ptr op);
int fmpq_poly_is_monic(const fmpq_poly_ptr op);
void fmpq_poly_monic(fmpq_poly_ptr rop, const fmpq_poly_ptr op);

///////////////////////////////////////////////////////////////////////////////
// Resultant

void fmpq_poly_resultant(mpq_t rop, const fmpq_poly_ptr a, const fmpq_poly_ptr b);
void fmpq_poly_discriminant(mpq_t disc, fmpq_poly_t a);

///////////////////////////////////////////////////////////////////////////////
// Composition

void fmpq_poly_compose(fmpq_poly_ptr rop, const fmpq_poly_ptr a, const fmpq_poly_ptr b);
void fmpq_poly_rescale(fmpq_poly_ptr rop, const fmpq_poly_ptr op, const mpq_t x);

///////////////////////////////////////////////////////////////////////////////
// Square-free

int fmpq_poly_is_squarefree(const fmpq_poly_ptr op);

///////////////////////////////////////////////////////////////////////////////
// Subpolynomials

void fmpq_poly_getslice(fmpq_poly_ptr rop, const fmpq_poly_ptr op, long i, long j);
void fmpq_poly_shift(fmpq_poly_ptr rop, const fmpq_poly_ptr op, long n);
void fmpq_poly_truncate(fmpq_poly_ptr rop, const fmpq_poly_ptr op, unsigned long n);
void fmpq_poly_reverse(fmpq_poly_ptr rop, const fmpq_poly_ptr op);

///////////////////////////////////////////////////////////////////////////////
// String conversion

int fmpq_poly_from_string(fmpq_poly_ptr rop, const char * s);
char * fmpq_poly_to_string(const fmpq_poly_ptr op);
char * fmpq_poly_to_string_pretty(const fmpq_poly_ptr op, const char * x);

#endif

